package com.example.e13roomshoppinglist

import androidx.lifecycle.ViewModel
import androidx.lifecycle.viewModelScope
import kotlinx.coroutines.DelicateCoroutinesApi
import kotlinx.coroutines.ExperimentalCoroutinesApi
import kotlinx.coroutines.GlobalScope
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.flow.SharingStarted
import kotlinx.coroutines.flow.combine
import kotlinx.coroutines.flow.flatMapLatest
import kotlinx.coroutines.flow.stateIn
import kotlinx.coroutines.flow.update
import kotlinx.coroutines.launch


enum class SortType{
    None,
    NameASC,
    NameDESC,
    PriceASC,
    PriceDESC,
    CountASC,
    CountDESC,
}

sealed interface ShoppingListEvent{
    data object OpenAddingNewShoppingListItemDialog: ShoppingListEvent
    data object CloseAddingNewShoppingListItemDialog: ShoppingListEvent
    data class SetName(val name: String): ShoppingListEvent
    data class SetPrice(val price: String): ShoppingListEvent
    data class SetCount(val count: String): ShoppingListEvent
    data object InsertNewShoppingListItemToDb: ShoppingListEvent

    data class OpenEditingShoppingListItem(val shoppingListItem: ShoppingListItem): ShoppingListEvent
    data object CloseEditingShoppingListItem: ShoppingListEvent
    data object UpdateCurrentlyEditingShoppingListItem: ShoppingListEvent
    data class DeleteShoppingListItem(val shoppingListItem: ShoppingListItem): ShoppingListEvent

    data class SetSortType(val sortType: SortType): ShoppingListEvent
}

data class NewShoppingListItem(
    val name: String = "",
    val price: String = "",
    val count: String = ""
)

data class ShoppingListState(
    val shoppingListItems: List<ShoppingListItem> = emptyList(),

    val isAddingNewShoppingListItem: Boolean = false,
    val newShoppingListItem: NewShoppingListItem = NewShoppingListItem(),

    val isEditingShoppingListItem: Boolean = false,
    val currentlyEditingShoppingListItem: ShoppingListItem? = null,

    val sortType: SortType = SortType.None
)


class ShoppingListViewModel(private val dao: ShoppingListDao): ViewModel() {
    private val _state = MutableStateFlow(ShoppingListState())
    private val _sortType = MutableStateFlow(SortType.None)
    @OptIn(ExperimentalCoroutinesApi::class)
    private val _shoppingListItems = _sortType.flatMapLatest { sortType ->
        when(sortType) {
            SortType.None -> dao.getAllShoppingListItems()
            SortType.NameASC -> dao.getShoppingListItemsOrderedByNameASC()
            SortType.NameDESC -> dao.getShoppingListItemsOrderedByNameDESC()
            SortType.PriceASC -> dao.getShoppingListItemsOrderedByPriceASC()
            SortType.PriceDESC -> dao.getShoppingListItemsOrderedByPriceDESC()
            SortType.CountASC -> dao.getShoppingListItemsOrderedByCountASC()
            SortType.CountDESC -> dao.getShoppingListItemsOrderedByCountDESC()
        }
    }.stateIn(viewModelScope, SharingStarted.WhileSubscribed(), emptyList())

    val state = combine(_state, _sortType, _shoppingListItems){
        state, sortType, shoppingListItems -> state.copy(shoppingListItems = shoppingListItems, sortType=sortType)
    }.stateIn(viewModelScope, SharingStarted.WhileSubscribed(5000), ShoppingListState())



    @OptIn(DelicateCoroutinesApi::class)
    fun onEvent(event: ShoppingListEvent){
        when(event) {
            is ShoppingListEvent.SetName -> {
                val newShoppingListItem = _state.value.newShoppingListItem
                _state.update { it.copy(newShoppingListItem = newShoppingListItem.copy(name = event.name)) }
            }
            is ShoppingListEvent.SetCount -> {
                val newShoppingListItem = _state.value.newShoppingListItem
                _state.update { it.copy(newShoppingListItem = newShoppingListItem.copy(count = event.count)) }
            }
            is ShoppingListEvent.SetPrice -> {
                val newShoppingListItem = _state.value.newShoppingListItem
                _state.update { it.copy(newShoppingListItem = newShoppingListItem.copy(price = event.price)) }
            }

            ShoppingListEvent.OpenAddingNewShoppingListItemDialog -> {
                _state.update { it.copy(isAddingNewShoppingListItem = true) }
            }
            ShoppingListEvent.InsertNewShoppingListItemToDb -> {
                val newShoppingListItem = _state.value.newShoppingListItem
                GlobalScope.launch {
                    val price = newShoppingListItem.price.toFloatOrNull()
                    val count = newShoppingListItem.count.toIntOrNull()
                    if (price===null || count===null) return@launch

                    dao.upsertShoppingListItem(ShoppingListItem(
                        name = newShoppingListItem.name,
                        price = price,
                        count = count))
                }
                _state.update { it.copy(newShoppingListItem= NewShoppingListItem()) }
            }
            ShoppingListEvent.CloseAddingNewShoppingListItemDialog -> {
                _state.update { it.copy(isAddingNewShoppingListItem = false) }
            }

            is ShoppingListEvent.OpenEditingShoppingListItem -> {
                _state.update {
                    val shoppingListItem = event.shoppingListItem

                    it.copy(
                    isEditingShoppingListItem = true,
                    currentlyEditingShoppingListItem = shoppingListItem,
                    newShoppingListItem = NewShoppingListItem(
                        name = shoppingListItem.name,
                        price = shoppingListItem.price.toString(),
                        count = shoppingListItem.count.toString()
                    ))
                }
            }
            ShoppingListEvent.UpdateCurrentlyEditingShoppingListItem -> {
                val currentlyEditingShoppingListItem = _state.value.currentlyEditingShoppingListItem
                if (currentlyEditingShoppingListItem===null) return
                val modifiedShoppingListItemAttrs = _state.value.newShoppingListItem
                val price = modifiedShoppingListItemAttrs.price.toFloatOrNull()
                val count = modifiedShoppingListItemAttrs.count.toIntOrNull()
                if (price===null || count===null) return

                val updatedShoppingListItem = currentlyEditingShoppingListItem.copy(
                    name = modifiedShoppingListItemAttrs.name,
                    price = price,
                    count = count
                )

                GlobalScope.launch { dao.upsertShoppingListItem(updatedShoppingListItem) }

                _state.update {
                    it.copy(
                        currentlyEditingShoppingListItem = null,
                        newShoppingListItem = NewShoppingListItem(),
                        isEditingShoppingListItem = false)
                }
            }
            ShoppingListEvent.CloseEditingShoppingListItem -> {
                _state.update {
                    it.copy(
                    isEditingShoppingListItem = false,
                    currentlyEditingShoppingListItem = null,
                    newShoppingListItem = NewShoppingListItem())
                }
            }
            is ShoppingListEvent.DeleteShoppingListItem -> {
                GlobalScope.launch {
                    dao.deleteShoppingListItem(event.shoppingListItem)
                }
            }

            is ShoppingListEvent.SetSortType -> {
                _sortType.update { event.sortType }
            }
        }
    }
}
